/* --------------------------------------------------------------
    Integral arithmetical calculator.

    http://www.imach.uran.ru/cbignum

    Copyright 2002-2009 by Raul N.Shakirov, IMach of RAS(UB).
    All Rights Reserved.

    Permission has been granted to copy, distribute and modify
    software in any context without fee, including a commercial
    application, provided that the aforesaid copyright statement
    is present here as well as exhaustive description of changes.

    THE SOFTWARE IS DISTRIBUTED "AS IS". NO WARRANTY OF ANY KIND
    IS EXPRESSED OR IMPLIED. YOU USE AT YOUR OWN RISK. THE AUTHOR
    WILL NOT BE LIABLE FOR DATA LOSS, DAMAGES, LOSS OF PROFITS OR
    ANY OTHER KIND OF LOSS WHILE USING OR MISUSING THIS SOFTWARE.
-------------------------------------------------------------- */
#include <assert.h>
#include <ctype.h>
#include <stdlib.h>
#include <string.h>
#include "Cbignum.h"
#include "Cbignumf.h"
#include "Cbignums.h"
#include "Gettimer.h"
#include "Random3.h"

#ifdef  _CIOS_STDSTREAM
#include <fstream>
#else //_CIOS_STDSTREAM
#include <fstream.h>
#endif//_CIOS_STDSTREAM

#ifdef  _CIOS_STDSTREAM
using namespace std;
#endif//_CIOS_STDSTREAM

static void GenerateRandom (cBigNumber& a, size_t radix, size_t maxexp = 0);
static void GenerateRandom (long a, size_t radix, size_t maxexp = 0);
static void GenerateRandomExp (size_t maxexp = 0);
static void ShowStat (cBigNumber& a, cBigNumber& na);
static void ShowCycles (cBigNumber& nt);

long SetTimer (long &timer)
{
  long temp = GetTimer(0);
  long time = temp - timer;
  timer = temp;

  return (time >= 0? time: 0L);
}

int GetNumberOp (_CIOSTD istream& is, cBigNumber& a)
{
  is >> a >> ws;

  int c = is.get();
  switch (c)
  {
    case 'R': case 'r':
      long la = a.tolong();
      a.setrandom (lRand3, (la < 0? -la: la));
      if (la < 0) a.neg();
      is >> ws; c = is.get();
      break;
  }
  return c;
}

int WriteNumber (_CIOSTD ostream& os, const cBigNumber& b, int nCheck)
{
  assert (&os);
  assert (&b);
  long     flags = (&os? os.flags(): 0);

  unsigned radix = ((flags & _CIOSTD ios::hex)? 16:
                    (flags & _CIOSTD ios::oct)? 8: 10);
  unsigned width = 0;
  long     fill  = 0;
  if (flags & _CIOSTD ios::internal)
  {
    width = (unsigned) os.width();
    fill  = (long)     os.fill();
  }
#ifdef  _CBIGNUM_UNSIGN_OCT_HEX
  if (radix != 10) fill |= cBigNumber_unsign;
#endif//_CBIGNUM_UNSIGN_OCT_HEX
  if (flags & _CIOSTD ios::uppercase) fill |= cBigNumber_uppercase;
  if (flags & _CIOSTD ios::showbase)  fill |= cBigNumber_showbase;
  if (flags & _CIOSTD ios::showpos)   fill |= cBigNumber_showpos;

  cBigString str;
  os << b.toa (str, radix, width, fill) << endl;

  if (nCheck)
#ifdef  _CBIGNUM_UNSIGN_OCT_HEX
  if (radix != 10 || b >= 0)
#endif//_CBIGNUM_UNSIGN_OCT_HEX
  {
    cBigNumber c (str, (flags & _CIOSTD ios::showbase)? 0: radix);
    if (c != b)
    {
      os << "NO MATCH IN OUTPUT: DIFFERENCE " << (c -= b) << endl;
      cerr << "FAIL!" << endl;
      return 1;
    }
  }
  return 0;
}

inline size_t size_max (size_t n1, size_t n2)
{
  return ((n1 > n2)? n1 : n2);
}

inline size_t size_min (size_t n1, size_t n2)
{
  return ((n1 < n2)? n1 : n2);
}

cBigNumber& csetmul (const cBigNumber& a, const cBigNumber& b, cBigNumber& c)
{
  size_t la = a.length(), lb = b.length();
  c.checkexpand (la + lb + 2);
  if (la <= lb)
        cBigNumberMul (EXPTRTYPE(a), EXPTRTYPE(b), EXPTRTYPE(c));
  else  cBigNumberMul (EXPTRTYPE(b), EXPTRTYPE(a), EXPTRTYPE(c));
  c.checkindex (c.length());
  return c;
}

cBigNumber& caddmul (const cBigNumber& a, const cBigNumber& b, cBigNumber& c)
{
  size_t la = a.length(), lb = b.length();
  c.checkexpand (size_max (c.length(), la + lb + 1) + 1);
  if (la <= lb)
        cBigNumberMAddMul (EXPTRTYPE(a), EXPTRTYPE(b), EXPTRTYPE(c));
  else  cBigNumberMAddMul (EXPTRTYPE(b), EXPTRTYPE(a), EXPTRTYPE(c));
  c.checkindex (c.length());
  return c;
}

cBigNumber& csubmul (const cBigNumber& a, const cBigNumber& b, cBigNumber& c)
{
  size_t la = a.length(), lb = b.length();
  c.checkexpand (size_max (c.length(), la + lb + 1) + 1);
  if (la <= lb)
        cBigNumberMSubMul (EXPTRTYPE(a), EXPTRTYPE(b), EXPTRTYPE(c));
  else  cBigNumberMSubMul (EXPTRTYPE(b), EXPTRTYPE(a), EXPTRTYPE(c));
  c.checkindex (c.length());
  return c;
}

const char szID[] =
    "Big integer calculator. Version 1.2b public of "__DATE__"\n"
    "Copyright 2002-2009 by Dr. Raul N.Shakirov, IMach of RAS(UB)\n"
    ;
const char szHelp[] =
    "Call:   Arifexp [Options] [Infile]\n"
    "Options:\n"
    "-idle   idle priority\n"
    "-high   high priority\n"
    "-hex    hexadecimal input/output\n"
    "-hexi   hexadecimal input\n"
    "-hexo   hexadecimal output\n"
    "-div0   allow division by 0\n"
    "-size   show high word and size on input numbers\n"
    "-time   show computing time estimation in Athlon cycles\n"
    "-v      write expression before the result\n"
    "-rep n  repeat operation n times\n"
    "-check  check results of operations, if possible\n"
    "-rand   generate random numbers instead of computing:\n"
    "        positive number of length n instead of value n\n"
    "        negative number of length n instead of value -n\n"
    "        number -8,-4,-2,-1,0,1,2,4,8 instead of 0 if option -rand 0\n"
    "        but keep degree of shift and power, unless option -rand all\n"
    "-exp n  append up to n trailing zeroes to random numbers\n"
    "-help   show this help and exit\n"
    "\n"
    "Site:   http://www.imach.uran.ru/cbignum\n"
    ;

int main (int argc, char** argv)
{
  int radix    = 10;
  int nTime    = 0;
  int nSize    = 0;
  int nVerbose = 0;
  int nRepeat  = 1;
  int nCheck   = 0;
  int maxexp   = 0;

  int  retcode = 0;     // Error code
  long total_r = 0;     // Time to read
  long total_c = 0;     // Time to calculate
  long total_w = 0;     // Time to write

  cout << dec;
  cerr << szID << endl;

  const char *pszPriority = "";
  const char *pszGenerate = "";
  const char *pszAll      = "";
  int nRand0   = 0;
  int nHexi    = 0;

  // Initialize

  cBigNumberMessages = cTTY_StandardOutput;
  cBigNumberProgress = cTTY_ProgressOutput;
  RRand3();

  // Look for options.

  for (; argc > 1 && argv [1] [0] == '-'; --argc, ++argv)
  {
    if (argv [1] [1] == '-') ++(argv [1]);

    if (strcmp (argv [1], "-idle") == 0)
    {
      if (SetIdlePriority()) pszPriority = "Idle priority\n\n";
      continue;
    }
    if (strcmp (argv [1], "-high") == 0)
    {
      if (SetHighPriority()) pszPriority = "High priority\n\n";
      continue;
    }
    if (strcmp (argv [1], "-hex") == 0)
    {
      nHexi = 1;
      cout << hex; radix = 16;
      continue;
    }
    if (strcmp (argv [1], "-hexi") == 0)
    {
      nHexi = 1;
      continue;
    }
    if (strcmp (argv [1], "-hexo") == 0)
    {
      cout << hex; radix = 16;
      continue;
    }
    if (strcmp (argv [1], "-div0") == 0)
    {
      cBigNumber::maskdiv0 (1);
      continue;
    }
    if (strcmp (argv [1], "-size") == 0)
    {
      nSize = 1;
      continue;
    }
    if (strcmp (argv [1], "-time") == 0)
    {
      nTime = 1;
      continue;
    }
    if (strcmp (argv [1], "-v") == 0)
    {
      nVerbose = 1;
      continue;
    }
    if (strcmp (argv [1], "-rep") == 0 && argc > 2)
    {
      nRepeat = atoi (argv [2]);
      --argc, ++argv;
      continue;
    }
    if (strcmp (argv [1], "-check") == 0)
    {
      nCheck = 1;
      continue;
    }
    if (strcmp (argv [1], "-rand") == 0)
    {
      pszGenerate = "Generating random numbers\n\n";
      if (argc > 2 && strcmp (argv [2], "all") == 0)
      {
        pszAll = argv[2];
        --argc, ++argv;
      }
      if (argc > 2 && strcmp (argv [2], "0") == 0)
      {
        nRand0 = 1;
        --argc, ++argv;
      }
      continue;
    }
    if (strcmp (argv [1], "-exp") == 0 && argc > 2)
    {
      maxexp = atoi (argv [2]);
      --argc, ++argv;
      continue;
    }
    if (strcmp (argv [1], "-help") == 0 ||
        strcmp (argv [1], "-h")    == 0 ||
        strcmp (argv [1], "-?")    == 0)
    {
      cout << szHelp;
      return 0;
    }

    // Unrecognized option.

    cerr << "Unrecognized option " << argv [1];
    cerr << " (type -help to view list of options)" << endl;
    return 255;
  }

  // Open input file

  istream *pin = &cin;
  ifstream fin;
  if (argc > 1)
  {
    fin.open (argv [1]);
    if (!fin)
    {
       cerr << "Can't read file " << argv [1] << endl;
       return 255;
    }
    pin = &fin;
  }

  // Set input mode

  if (nHexi) *pin >> hex; else *pin >> dec;

  // Report process mode

  cerr << pszPriority;
  cerr << pszGenerate;

  // Allocate big numbers (must be initialized within cycle before use).

  cBigNumber a;         // Operand 1
  cBigNumber b;         // Operand 2
  cBigNumber c;         // Operand 3
  cBigNumber null;      // Null number

  cBigNumber na;        // Number of bits in a
  cBigNumber nb;        // Number of bits in b
  cBigNumber nc;        // Number of bits in c
  cBigNumber nt;        // Estimated number of CPU cycles to input/calculate/output

  cBigNumber x;         // Result 1
  cBigNumber y;         // Result 2
  cBigNumber p;         // Number to compare with result
  cBigNumber q;         // Number to compare with result
  cBigNumber v;         // Number to vote for correct result
  cBigNumber cmp;       // Operand to compare

  // Process expressions.

  int ch;
  do
  {
    // Get expression.

    if (argc > 1)
    {
      ch = fin.get();
      if (ch == EOF) break;
      fin.putback ((char)ch);
    }
    else
    {
      cerr << "Please, type expression "\
            "a+-*/%\\|&^?b=, a+-b*c=, a\\b%c=, a<<>>b=, 2Vb, aR=\n"\
            "and press Enter or press Enter to exit\n";

      ch = cin.get();
      if (ch == EOF || ch == '\r' || ch == '\n') break;
      cin.putback ((char)ch);
    }

    long timer = GetTimer (0);

    a = 0;              // Operand 1
    b = 0;              // Operand 2
    c = 0;              // Operand 3
    null = 0;           // Null number

    int pp = ch;        // First symbol
    int op = 0, md = 0; // Operators
    unsigned error = 0; // Error mask

    unsigned alt = 1;   // Use alternative operation in check mode
    unsigned rev = 1;   // Use reverse operation in check mode
    unsigned arb = 0;   // Use simplified arbitrage algorithm in check mode

    ch = GetNumberOp (*pin, a);

    switch (ch)
    {
      case '<': case '>': op = ch; *pin >> ws; ch = pin->get();
                          if (op != ch) op = ch = ' ';  // Error
    }

    if (ch != EOF && ch != '=')
    {
      op = ch; ch = GetNumberOp (*pin, b);
    }

    if (ch != EOF && ch != '=')
    {
      md = ch; ch = GetNumberOp (*pin, c);
    }

    // Check operands to be correct and calculate number of cycles.

    na.setbits (a);     // Number of bits in a
    nb.setbits (b);     // Number of bits in b
    nc.setbits (c);     // Number of bits in c
    nt = 0;             // Estimated number of CPU cycles to input/calculate/output

    #ifdef  _CBIGNUM_HARDWARE_CBR
    #define DADD 15
    #define DDIV 25
    #define DSQR 120
    #define DOUT 75
    #else
    #define DADD 5
    #define DDIV 8
    #define DSQR 40
    #define DOUT 25
    #endif

    #ifdef  _CBIGNUM_HARDWARE_MUL
    #define BMUL 31
    #define DMUL 200
    #define DINP 400
    #define DPWM 17
    #else
    #ifdef  _CBIGNUM_HARDWARE_CBR
    #define BMUL 0
    #define DMUL 45
    #define DINP 100
    #define DPWM 15
    #else
    #define BMUL 0
    #define DMUL 15
    #define DINP 33
    #define DPWM 5
    #endif
    #endif

    #define DMKR (60 * _CBNL_KARATSUBA_MIN)
    #define DIKR (1000 * _CBNL_KARATSUBA_MIN)

    if (op != 0)
    {
      if (md == 0)
      switch (op)
      {
        case '+':   if (pp == op && a == 0) { op = +1; a = b; break; }
                    if (nTime) { nt = (na > nb? na: nb); nt /= DADD; }
                    break;
        case '-':   if (pp == op && a == 0) { op = -1; a = b; break; }
                    if (nTime) { nt = (na > nb? na: nb); nt /= DADD; }
                    break;
        case '*':   if (a == 0 || b == 0) rev = 0;
                    if (nTime) {
                    #ifdef _CBIGNUM_KARATSUBA_MUL
                      nt.setbits ((na < nb? na: nb) / DMKR);
                      nt = (na | BMUL) * (nb | BMUL) *
                        cBigPow (3, nt) / (cBigPow (4, nt) * DMUL);
                    #else
                      nt = (na | BMUL) * (nb | BMUL) / DMUL;
                    #endif
                    }
                    arb = 1;
                    break;
        case '/':
        case '%':   alt = 0;
                    if (nTime) { nt = nb; if (na > nb) nt *= (na - nb);
                                 nt /= DDIV; }
                    break;
        case '\\':  switch (b.loword()) {
                      case 32: case 16: case 8: case 4: case 2: case 1:
                              if (b.length() == 1) break;
                      default: alt = 0; rev = 0;
                    }        break;
        case '<':            break;
        case '>':   rev = 0; break;
        case '|':   rev = 0; break;
        case '&':   rev = 0; break;
        case '^':            break;
        case 'V':
        case 'v':   alt = 0;
                    if (nTime) nt = nb * nb / DSQR;
                    if (a == 2) break;
        case '?':   rev = 0; break;
        default:    error = 0x0001;
      }
      else if (md == '*')
      switch (op)
      {
        case '+':
        case '-':   if (b == 0 || c == 0) rev = 0;
                    if (nTime) {
                    #ifdef _CBIGNUM_KARATSUBA_MUL
                      nt.setbits ((na < nb? na: nb) / DMKR);
                      nt = (na | BMUL) * (nb | BMUL) *
                        cBigPow (3, nt) / (cBigPow (4, nt) * DMUL);
                    #else
                      nt = (na | BMUL) * (nb | BMUL) / DMUL;
                    #endif
                    }
                    arb = 1;
                    break;
        default:    error = 0x0001;
      }
      else if (md == '%')
      switch (op)
      {
        case '\\':  null = na; null *= b; null /= 8;
                    if (null > (EXSIZE_T_MAX & (size_t)0x00FFFFFFUL) / 2 ||
                        c == 0) alt = 0;
                    rev = 0;
                    if (nTime) nt = nb * nc * nc / DPWM;
                    break;
       default:     error = 0x0001;
      }
      else error = 0x0002;

      nt *= nRepeat;
    }
    else if (nTime)
    {
      #ifdef _CBIGNUM_KARATSUBA_MUL
        nt.setbits (na / DIKR);
        nt = (na * na * cBigPow (3, nt)) / (cBigPow (4, nt) * DINP);
      #else
        nt = na * na / DINP;
      #endif
    }

    // Accumulate time to read.

    total_r += SetTimer (timer);

    // Put error diagnostics

    while (ch != EOF && ch != '=')
    {
      error |= 0x4000; ch = pin->get();         // Skip until =
    }

    for (ch = pin->get(); ch != EOF && ch != '\n'; ch = pin->get())
    {
      error |= (isspace (ch) - 1) & 0x8000;     // Skip rest of line
    }

    if (error & 0x0001)
      cout << "Error: 1st operator is either missed or unknown" << endl;

    if (error & 0x0002)
      cout << "Error: 2nd operator is either missed or unknown" << endl;

    if (error & 0x4000)
      cout << "Error: excessive operators or operands before sign =" << endl;

    if (error & 0x8000)
      cout << "Error: misplaced operators or operands after sign =" << endl;

    if (error)
    {
      retcode |= 255;
      cout << endl;
      continue;
    }

    // Generate random numbers.

    if (*pszGenerate != 0)
    {
      switch (op)
      {
        case  +1:
        case  -1:   cout << (char) pp << (char) pp;
                    break;
        case 'V':
        case 'v':   cout << a << endl << (char) op;
                    break;
        default:    if (a != 0 || nRand0 == 0)
                         GenerateRandom (a, radix, maxexp);
                    else GenerateRandomExp (maxexp);
                    cout << endl;
                    if (op != 0) cout << (char) op;
      }

      if (op != 0)
      {
        switch (op)
        {
          case '<':
          case '>': cout << (char) op;
          case '\\':if (*pszAll == 0)
                    {
                       cout << endl << b;
                       break;
                    }
          default:  cout << endl;
                    if (b != 0 || nRand0 == 0)
                         GenerateRandom (b, radix, maxexp);
                    else GenerateRandomExp (maxexp);
        }
        if (md != 0)
        {
          cout << endl << (char) md << endl;
          if (c != 0 || nRand0 == 0)
               GenerateRandom (c, radix, maxexp);
          else GenerateRandomExp (maxexp);
        }
        cout << endl;
      }
      cout << '=' << endl;
      continue;
    }

    // Report size of numbers.

    if (nSize)
    {
      switch (op)
      {
        case  +1:   break;
        case  -1:   break;
        case 'V':   break;
        case 'v':   break;
        default:    ShowStat (a, na);
      }
      if (op != 0)  ShowStat (b, nb);
      if (md != 0)  ShowStat (c, nc);
    }

    // Report number of cycles for operation.

    if (nTime && nt > 0)
    {
      cerr << "Estimating " << setw (6);
      ShowCycles (nt);
      if (op == 0)
      {
        cerr << " for input " << setw (6);
        cerr << endl;
        cerr << "Estimating " << setw (6);
        ShowCycles (nt = (na * na) / DOUT);
        cerr << " for output";
      }
      cerr << endl;
    }

    // Write operands.

    if (nVerbose != 0)
    {
      switch (op)
      {
        case  +1:
        case  -1:   cout << (char) pp << (char) pp;
                    break;
        default:    retcode |= WriteNumber (cout, a, nCheck);
      }
      if (op != 0)
      {
        switch (op)
        {
          case '<':
          case '>': cout << (char) op;
          default:  cout << (char) op;
          case  +1:
          case  -1: cout << endl;
                    retcode |= WriteNumber (cout, b, nCheck);
        }
        if (md != 0)
        {
          cout << (char) md << endl;
          retcode |= WriteNumber (cout, c, nCheck);
        }
      }
      cout << '=' << endl;
    }

    // Main calculations

    if (op != 0)
    {
      // Accumulate time to write.

      total_w += SetTimer (timer);

      // Obtain the result.

      for (int n = 0;;)
      {
        if(!n)cerr << "Computing ";

        x = a;          // Result 1
        y = 0;          // Result 2

        if (md == 0)
        switch (op)
        {
          case  +1: if (!n) cerr <<"++b";       ++x;                break;
          case  -1: if (!n) cerr <<"--b";       --x;                break;
          case '+': if (!n) cerr <<"a+b";       x += b;             break;
          case '-': if (!n) cerr <<"a-b";       x -= b;             break;
          case '*': if (!n) cerr <<"a*b";       x *= b;             break;
          case '/': if (!n) cerr <<"a/b,a%b";   x.setdivmod (y = a, b);
                                                                    break;
          case '%': if (!n) cerr <<"a%b";       x %= b;             break;
          case '\\':if (!n) cerr <<"a\\b (pow)";x.pow (b);          break;
          case '<': if (!n) cerr <<"a<<b";      x <<= b;            break;
          case '>': if (!n) cerr <<"a>>b";      x >>= b;            break;
          case '|': if (!n) cerr <<"a|b (or)";  x |= b;             break;
          case '&': if (!n) cerr <<"a&b (and)"; x &= b;             break;
          case '^': if (!n) cerr <<"a^b (xor)"; x ^= b;             break;
          case 'V':
          case 'v': if (!n) cerr <<"sqrt(b),b-sqrt(b)*sqrt(b)";
                                                x.setsqrtrm (y = b);break;
          case '?': if (!n) cerr <<"a?b (cmp)"; x = x.comp(b);      break;
        }
        else if (md == '*')
        switch (op)
        {
          case '+': if (!n) cerr <<"a+b*c";     x.addmul (b,c);     break;
          case '-': if (!n) cerr <<"a-b*c";     x.submul (b,c);     break;
        }
        else if (md == '%')
        switch (op)
        {
          case '\\':if (!n) cerr <<"a\\b%c";    x.powmod (b,c);     break;
        }

        if (!n && nRepeat > 1) cerr <<" "<< nRepeat << " times";
        if (++n >= nRepeat) { cerr << endl; break; }
      }

      // Accumulate time to calculate.

      total_c += SetTimer (timer);

      // Write the result.

      retcode |= WriteNumber (cout, x, nCheck);
      switch (op)
      {
        case '/':
        case 'V':
        case 'v': retcode |= WriteNumber (cout, y, nCheck);
                  break;
      }
    }
    else // Just write the operand
    {
      retcode |= WriteNumber (cout, a, nCheck);
    }

    // Accumulate time to write.

    total_w += SetTimer (timer);

    // Comparing by alternative operation.

    if (nCheck != 0 && op != 0 && alt != 0)
    {
      p = 0;            // Number to compare with result
      q = 0;            // Number to compare with result

      cerr << "Computing ";

      if (md == 0)
      switch (op)
      {
        case +1:  cerr <<"b+1";         p = b + 1;              break;
        case -1:  cerr <<"b-1";         p = b - 1;              break;
        case '+': cerr <<"a+b";         p = a + b;              break;
        case '-': cerr <<"a-b";         p = a - b;              break;
        case '*': cerr <<"b*a";         csetmul (a,b,p);        break;
        case '/': cerr <<"a/b,a%b";     p = a / b; q = a % b;   break;
        case '%': cerr <<"a%b";         p = a % b;              break;
        case '\\':cerr <<"a";           p = a;
                  switch (b.loword()) {
          case 32:cerr << "\\2";        p.pow2();
          case 16:cerr << "\\2";        p.pow2();
          case 8: cerr << "\\2";        p.pow2();
          case 4: cerr << "\\2";        p.pow2();
          case 2: cerr << "\\2";        p.pow2();
                  }                                             break;
        case '<': if (b >= 0)
                        { cerr <<"a*2\\b";  p = a * cBigPow (2,b);  }
                  else  { cerr <<"floor(a/2\\-b)";
                                        p = cBigPow (2,-b);
                           if (a >= 0)  p = a / p;
                           else         p = (a - p + 1) / p; }  break;
        case '>': if (b < 0)
                        { cerr <<"a*2\\-b"; p = a * cBigPow (2,-b); }
                  else  { cerr <<"floor(a/2\\b)";
                                        p = cBigPow (2,b);
                           if (a >= 0)  p = a / p;
                           else         p = (a - p + 1) / p; }  break;
        case '|': cerr <<"~(~a&~b)";    p = ~(~a & ~b);         break;
        case '&': cerr <<"~(~a|~b)";    p = ~(~a | ~b);         break;
        case '^': cerr <<"~a&b|a&~b";   p = (~a & b) | (a & ~b);break;
        case '?': cerr <<"a-b?";        p = a; p -= b; p = p.comp0();
                                                                break;
      }
      else if (md == '*')
      switch (op)
      {
        case '+': cerr <<"a+c*b";       caddmul (b,c,p=a);      break;
        case '-': cerr <<"a-c*b";       csubmul (b,c,p=a);      break;
      }
      else if (md == '%')
      switch (op)
      {
        case '\\':cerr <<"(a\\b)%c";    p = cBigPow(a,b) % c;   break;
      }
      cerr << endl;

      total_c += SetTimer (timer);

      retcode |= WriteNumber (cout, p, nCheck);
      switch (op)
      {
        case '/':
        case 'V':
        case 'v':   retcode |= WriteNumber (cout, q, nCheck);
                    break;
      }

      total_w += SetTimer (timer);

      cerr << "Comparing... ";

      if (x == p && y == q)
      {
        cerr << "OK";
      }
      else
      {
        cout << "NO MATCH" << endl;
        cerr << "FAIL!";
        retcode |= 1;
      }
      cerr << endl;
    }

    // Comparing by arbitrage operation.

    if (nCheck != 0 && op != 0 && alt != 0 && arb != 0)
    {
      v = 0;            // Number to vote for result

      cerr << "Computing ";

      if (md == 0)
      switch (op)
      {
        case '*': cerr <<"a*b (tab)";   if (b >= 0) v.addmultab (a.tab(), b);
                                               else v.submultab (a.tab(), -b);
                                        break;
      }
      else if (md == '*')
      switch (op)
      {
        case '+': cerr <<"a+b*c (tab)"; v = a;
                                        if (c >= 0) v.addmultab (b.tab(), c);
                                               else v.submultab (b.tab(), -c);
                                        break;
        case '-': cerr <<"a-b*c (tab)"; v = a;
                                        if (c >= 0) v.submultab (b.tab(), c);
                                               else v.addmultab (b.tab(), -c);
                                        break;
      }

      cerr << endl;

      total_c += SetTimer (timer);

      retcode |= WriteNumber (cout, v, nCheck);

      total_w += SetTimer (timer);

      cerr << "Comparing... ";

      int f = 0;
      if (x == v) f |= 1;
      if (p == v) f |= 2;
      if (f == 2) x = p;

      if (f)
      {
        cerr << "OK";
        if (f != 3)
        {
          cerr << f << endl;
          cout << "MATCH WITH RESULT " << f << endl;
          retcode |= 1;
        }
        else
        {
          cerr << endl;
        }
      }
      else
      {
        cerr << "FAIL!" << endl;
        cout << "NO MATCH" << endl;
        retcode |= 1;
      }
    }

    // Comparing by reverse operation.

    if (nCheck != 0 && op != 0 && rev != 0)
    {
      p = x;            // Number to compare with operand
      q = 0;            // Number to compare with 0
      cmp = a;          // Operand to compare

      cerr << "Computing ";

      if (md == 0)
      switch (op)
      {
        case  +1: cerr << "--++b";      --p;                break;
        case  -1: cerr << "++--b";      ++p;                break;
        case '+': cerr << "a+b-b";      p -= b;             break;
        case '-': cerr << "a-b+b";      p += b;             break;
        case '*': cerr << "a*b/b";      p.setdivmod (q = p, b);
                                                            break;
        case '/': cerr << "a/b*b+a%b";  p *= b; p += y;     break;
        case '%': cerr << "a%b+a/b*b";  p += (a/b)*b;       break;
        case '\\':if ((b.loword() & 1) == 0 && cmp < 0) cmp.neg();
                  switch (b.loword()) {
          case 32:cerr << "sqrt(a\\32)";p.setsqrtrm (q = p);
                                                     if (q) break;
          case 16:cerr << "sqrt(a\\16)";p.setsqrtrm (q = p);
                                                     if (q) break;
          case 8: cerr << "sqrt(a\\8)"; p.setsqrtrm (q = p);
                                                     if (q) break;
          case 4: cerr << "sqrt(a\\4)"; p.setsqrtrm (q = p);
                                                     if (q) break;
          case 2: cerr << "sqrt(a\\2)"; p.setsqrtrm (q = p);
                  }                                         break;
        case '<': cerr << "(a<<b)>>b";  p >>= b;            break;
        case '^': cerr << "a^b^b";      p ^= b;             break;
        case 'V':
        case 'v': cerr << "sqrt(b)*sqrt(b)+(b-sqrt(b)*sqrt(b))";
                               cmp = b; p.pow2(); p += y;   break;
      }
      else if (md == '*')
      switch (op)
      {
        case '+': cerr << "(a+b*c-a)/c";p = p - a;
                                        p.setdivmod (q = p, c);
                                        cmp = b;            break;
        case '-': cerr << "(a-a-b*c)/c";p = a - p;
                                        p.setdivmod (q = p, c);
                                        cmp = b;            break;
      }
      cerr << endl;

      total_c += SetTimer (timer);

      retcode |= WriteNumber (cout, p, nCheck);
      if (md == 0)
      switch (op)
      {
        case '*':
        case '\\':  retcode |= WriteNumber (cout, q, nCheck);
      }
      else if (md == '*')
      switch (op)
      {
        case '+':
        case '-':   retcode |= WriteNumber (cout, q, nCheck);
      }

      total_w += SetTimer (timer);

      cerr << "Comparing... ";

      if (cmp == p && q == 0)
      {
        cerr << "OK";
      }
      else
      {
        cout << "NO MATCH" << endl;
        cerr << "FAIL!";
        retcode |= 1;
      }
      cerr << endl;
    }

    cout << endl;
  }
  while (ch != EOF);

  cout << dec;

  if (*pszGenerate == 0)
  {
    long total = 0;
    cout << "Reading    " << setw (6) << total_r << " ms\n"; total += total_r;
    cout << "Computing  " << setw (6) << total_c << " ms";   total += total_c;
    if (nRepeat > 1)
      cout << " (" << ((total_c * 2 + nRepeat) / (nRepeat * 2)) << " ms "
           << nRepeat << " times)";
    cout << "\n";
    cout << "Writing    " << setw (6) << total_w << " ms\n"; total += total_w;
    cout << "Total      " << setw (6) << total   << " ms\n";
  }

  return retcode;
}

static void GenerateRandom (cBigNumber& n, size_t radix, size_t maxexp)
{
  GenerateRandom (n.tolong(), radix, maxexp);
}

static void GenerateRandom (long n, size_t radix, size_t maxexp)
{
  assert (radix > 1);

  if (n < 0) { cout << '-'; n = -n; }

  cerr << "Generating random number of length "<< n;
  if (maxexp != 0)
  {
    maxexp = (int) (dRand3() * maxexp);
    cerr << " and exponent " << (long)maxexp;
  }
  cerr << endl;

  if (n > 0)
  {
    cout << (int) (1 + (dRand3() * (radix - 1)));
    while (--n != 0) cout << (int) (dRand3() * radix);
    for (; maxexp > 0; --maxexp) cout << 0;
  }
  else
  {
    cout << n;
  }
}

static int ExpSet[]={-8,-4,-2,-1,0,1,2,4,8};
static void GenerateRandomExp (size_t maxexp)
{
  cerr << "Generating random number of set -8,-4,-2,-1,0,1,2,4,8";
  if (maxexp != 0)
  {
    maxexp = (int) (dRand3() * maxexp);
    cerr << " and exponent " << (long)maxexp;
  }
  cerr << endl;

  int n = ExpSet [(int) (dRand3() * 9)];
  cout << n;
  if (n != 0) for (; maxexp > 0; --maxexp) cout << 0;
}

static void ShowStat (cBigNumber& a, cBigNumber& na)
{
  cerr << (const char*) cHexDump (a.hiword());
  if (a.length() > 1)
  {
    cHexDump a2 (a.code() [a.length() - 2]);
    ((char*) (const char*) a2) [2] = 0;
    cerr <<" "<< (const char*) a2 <<"..";
  }
  else
    cerr <<"     ";

  cerr <<": "<< setw (7) << na <<" bit" << (na != 1?"s":" ");
  size_t la = a.words();
  cerr <<" in "<< setw (5) << (long)la <<" word" << ("s" + (la == 1));
  cerr << endl;
  //a.dump();
  //a.info();
}

static void ShowCycles (cBigNumber& nt)
{
  if (nt < 1000000L)
    cerr << ((nt + 499) / 1000) << 'K';
  else
    cerr << ((nt + 499999L) / 1000000L) << 'M';
  cerr << " Athlon cycles";
}
